import json
import logging
from typing import Text, Dict, Any

from aiohttp import ClientError
from sanic import Blueprint, response
from sanic.request import Request

from rasa.nlu.training_data.loading import RASA, MARKDOWN
import ruamel.yaml as yaml
from rasax.community import utils
from rasax.community.api.blueprints.models import _model_service
from rasax.community.api.decorators import (
    rasa_x_scoped,
    inject_rasa_x_user,
    validate_schema,
)
from rasax.community.services.data_service import DataService
from rasax.community.services.intent_service import INTENT_MAPPED_TO_KEY
from rasax.community.services.intent_service import IntentService
from rasax.community.services.logs_service import LogsService
from rasax.community.services.evaluation_service import EvaluationService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.stack_service import StackService
from rasax.community.services.stack_service import nlu_format
from rasax.community.utils import dump_yaml, error, get_text_hash, load_yaml
from rasa.nlu.config import InvalidConfigError

logger = logging.getLogger(__name__)


def _stack_service(
    request: Request, project_id: Text, default_environment: Text
) -> StackService:
    settings_service = SettingsService(request["db_session"])

    environment = utils.deployment_environment_from_request(
        request, default_environment
    )
    service = settings_service.get_stack_service(environment, project_id)
    if not service:
        error(
            404,
            "ServiceNotFound",
            "Service for requested environment '{}' not found.".format(environment),
        )

    return service


def _log_service(request: Request) -> LogsService:
    return LogsService(request["db_session"])


def _data_service(request: Request) -> DataService:
    return DataService(request["db_session"])


def _evaluation_service(
    request: Request, project_id: Text, default_environment: Text
) -> EvaluationService:
    service = _stack_service(request, project_id, default_environment)
    return EvaluationService(service, request["db_session"])


async def _create_message_log_from_query(
    request: Request, project_id: Text, query: Text
) -> Dict[Text, Any]:

    stack_service = _stack_service(request, project_id, "worker")
    stack_service_has_model = await stack_service.has_active_model()

    if stack_service_has_model:
        parse_data = await stack_service.parse(query)
    else:
        # We don't have an active model to create an initial guess of the intent and
        # entities. We still want to be able to create a log, and will use parse data
        # that only contains the query under the `text` key.
        parse_data = {"text": query}

    logs_service = _log_service(request)
    log_hash = get_text_hash(query)
    existing_log = logs_service.get_log_by_hash(log_hash)

    if existing_log:
        existing_log_dict = existing_log.as_dict()
        return logs_service.replace_log(
            existing_log_dict, parse_data, created_from_model=stack_service_has_model
        )
    else:
        return logs_service.create_log_from_parse_data(
            parse_data, created_from_model=stack_service_has_model
        )


def blueprint():
    nlu_endpoints = Blueprint("nlu_endpoints")

    @nlu_endpoints.route("/projects/<project_id>/logs", methods=["GET", "HEAD"])
    @rasa_x_scoped("logs.list")
    async def suggestions(request, project_id):
        limit = int(utils.default_arg(request, "limit", 0)) or None
        offset = int(utils.default_arg(request, "offset", 0)) or None
        text_query = utils.default_arg(request, "q", None)
        intent_query = utils.default_arg(request, "intent", None)
        fields = utils.fields_arg(
            request, {"user_input.intent.name", "user_input.text", "id"}
        )
        distinct = utils.bool_arg(request, "distinct", True)
        suggested, total_suggestions = _log_service(request).get_suggestions(
            text_query, intent_query, fields, limit, offset, distinct
        )

        return response.json(suggested, headers={"X-Total-Count": total_suggestions})

    # noinspection PyUnusedLocal
    @nlu_endpoints.route("/projects/<project_id>/logs/<log_id>", methods=["DELETE"])
    @rasa_x_scoped("logs.delete")
    @inject_rasa_x_user()
    async def archive_one(request, project_id, log_id, user=None):
        success = _log_service(request).archive(log_id)
        if not success:
            return error(
                400,
                "ArchiveLogFailed",
                "Failed to archive log with log_id {}".format(log_id),
            )
        return response.text("", 204)

    @nlu_endpoints.route("/projects/<project_id>/logs/<_hash>", methods=["GET", "HEAD"])
    @rasa_x_scoped("logs.get")
    async def get_log_by_hash(request, project_id, _hash):
        log = _log_service(request).get_log_by_hash(_hash)
        if not log:
            return error(
                400, "NluLogError", "Log with hash {} could not be found".format(_hash)
            )

        return response.json(log.as_dict())

    @nlu_endpoints.route("/projects/<project_id>/logs", methods=["POST"])
    @rasa_x_scoped("logs.create")
    async def add_log(request, project_id):
        query = utils.default_arg(request, "q", "")
        # if no query text found, check for json content
        if not query:
            query = request.json
            if not utils.check_schema("log", query):
                return error(
                    400, "WrongSchema", "Please check the schema of your NLU query."
                )

        try:
            created_log = await _create_message_log_from_query(
                request, project_id, query
            )
            return response.json(created_log)
        except ClientError as e:
            logger.warning(
                "Parsing the message '{}' failed. Error: {}".format(query, e)
            )
            return error(
                404,
                "NluParseFailed",
                "Failed to parse NLU query ('{}').".format(query),
                details=e,
            )

    @nlu_endpoints.route("/projects/<project_id>/data", methods=["GET", "HEAD"])
    @rasa_x_scoped("examples.list")
    async def training_data(request, project_id):
        """Retrieve the training data for the project."""

        offset = int(utils.default_arg(request, "offset", 0)) or None
        limit = int(utils.default_arg(request, "limit", 0)) or None
        text_query = utils.default_arg(request, "q", None)
        intent_query = utils.default_arg(request, "intent", None)
        entity_query = utils.bool_arg(request, "entities", False)
        should_sort_by_id = utils.bool_arg(request, "sorted", True)
        fields = utils.fields_arg(
            request, {"intent", "text", "id", "entities", "entities.entity"}
        )
        distinct = utils.bool_arg(request, "distinct", True)
        data_query = _data_service(request).get_training_data(
            project_id=project_id,
            sort_by_id=should_sort_by_id,
            text_query=text_query,
            intent_query=intent_query,
            entity_query=entity_query,
            fields_query=fields,
            limit=limit,
            offset=offset,
            distinct=distinct,
        )

        return response.json(
            data_query.result, headers={"X-Total-Count": data_query.count}
        )

    # noinspection PyUnusedLocal
    @nlu_endpoints.route("/projects/<project_id>/data/<_hash>", methods=["GET", "HEAD"])
    @rasa_x_scoped("examples.get")
    async def training_example_by_hash(request, project_id, _hash):
        """Retrieve a training example by its hash."""
        example = _data_service(request).get_example_by_hash(project_id, _hash)
        if not example:
            return error(400, "TrainingExampleError", "Example could not be found.")

        return response.json(example)

    # noinspection PyUnusedLocal
    @nlu_endpoints.route("/projects/<project_id>/dataWarnings", methods=["GET", "HEAD"])
    @rasa_x_scoped("warnings.get")
    async def training_data_warnings(request, project_id):
        """Retrieve the training data for the project."""

        warnings = _data_service(request).get_training_data_warnings(
            project_id=project_id
        )

        return response.json(warnings)

    @nlu_endpoints.route("/projects/<project_id>/data.json", methods=["GET", "HEAD"])
    @rasa_x_scoped("bulkData.get", allow_api_token=True)
    async def training_data_as_file(request, project_id):
        """Download the training data for the project in Rasa NLU format."""
        data_service = _data_service(request)
        data = data_service.get_training_data(project_id).result
        regex_features = data_service.get_regex_features(project_id)
        lookup_tables = data_service.get_lookup_tables(project_id)
        entity_synonyms = data_service.get_entity_synonyms(project_id)

        content = nlu_format(data, regex_features, lookup_tables, entity_synonyms)

        return response.text(
            json.dumps(content, indent=4, ensure_ascii=False),
            content_type="application/json",
            headers={"Content-Disposition": "attachment;filename=nlu.train.json"},
        )

    @nlu_endpoints.route("/projects/<project_id>/data", methods=["POST"])
    @rasa_x_scoped("examples.create")
    @inject_rasa_x_user()
    @validate_schema("data")
    async def add_training_example(request, project_id, user=None):
        """Add a new training example to the project."""

        rjs = request.json
        data_service = _data_service(request)
        example_hash = get_text_hash(rjs.get("text"))
        existing_example = data_service.get_example_by_hash(project_id, example_hash)
        if existing_example:

            example_id = str(existing_example["id"])
            example = data_service.replace_example(user, project_id, rjs, example_id)

        else:
            example = data_service.save_example(user["username"], project_id, rjs)

        if not example:
            return error(400, "SaveExampleError", "Example could not be saved")

        return response.json(example)

    @nlu_endpoints.route("/projects/<project_id>/data/<example_id>", methods=["PUT"])
    @rasa_x_scoped("examples.update")
    @inject_rasa_x_user()
    @validate_schema("data")
    async def update_example(request, project_id, example_id, user=None):
        """Update an existing training example."""

        rjs = request.json
        data_service = _data_service(request)
        example_hash = get_text_hash(rjs.get("text"))
        existing_example = data_service.get_example_by_hash(project_id, example_hash)
        if existing_example and example_id != str(existing_example["id"]):
            return error(
                400, "ExampleUpdateFailed", "Example with this text already exists"
            )

        mapped_to = rjs.get("intent_mapped_to")
        if mapped_to:
            intent_service = IntentService(request["db_session"])
            intent = {INTENT_MAPPED_TO_KEY: mapped_to}
            intent_service.update_temporary_intent(rjs["intent"], intent, project_id)
            intent_service.add_example_to_temporary_intent(
                rjs[intent], example_hash, project_id
            )
        updated_example = data_service.replace_example(
            user, project_id, rjs, example_id
        )
        if not updated_example:
            return error(400, "ExampleUpdateFailed", "Example could not be updated")

        return response.json(updated_example)

    # noinspection PyUnusedLocal
    @nlu_endpoints.route("/projects/<project_id>/data/<example_id>", methods=["DELETE"])
    @rasa_x_scoped("examples.delete")
    @inject_rasa_x_user()
    async def training_data_delete(request, project_id, example_id, user=None):
        """Remove a training example from the project by id."""

        success = _data_service(request).delete_example(example_id)
        if success:
            return response.text("", 204)
        else:
            return error(
                404,
                "DeleteTrainingExampleFailed",
                "Training example with example_id '{}' could not be "
                "deleted".format(example_id),
            )

    @nlu_endpoints.route("/projects/<project_id>/data", methods=["PUT"])
    @rasa_x_scoped("bulkData.update")
    @inject_rasa_x_user()
    async def training_data_bulk(request, project_id, user=None):
        """Replace existing training samples with the posted data."""

        content_type = request.headers.get("Content-Type")
        try:
            data_service = _data_service(request)
            if content_type == "application/json":
                rjs = request.json
                data_service.replace_data(project_id, rjs, RASA, user["username"])
            elif content_type == "text/markdown":
                data = request.body
                data_service.replace_data(project_id, data, MARKDOWN, user["username"])

            return response.text("Bulk upload of training data successful.")
        except ValueError as e:
            logger.error(e)
            return error(
                400, "AddTrainingDataFailed", "Failed to add training data", details=e
            )

    # noinspection PyUnusedLocal
    @nlu_endpoints.route("/projects/<project_id>/entities", methods=["GET", "HEAD"])
    @rasa_x_scoped("entities.list")
    @inject_rasa_x_user()
    async def get_entities(request, project_id, user=None):
        """Fetches a list of unique entities present in training data."""

        entities = _data_service(request).get_entities(project_id)
        return response.json(entities, headers={"X-Total-Count": len(entities)})

    @nlu_endpoints.route("/projects/<project_id>/evaluations", methods=["GET", "HEAD"])
    @rasa_x_scoped("models.evaluations.list", allow_api_token=True)
    async def evaluations(request, project_id):
        """Fetches a list of Rasa NLU evaluations."""

        evaluation_service = _evaluation_service(request, project_id, "worker")
        results = evaluation_service.formatted_evaluations(project_id)
        return response.json(results, headers={"X-Total-Count": len(results)})

    @nlu_endpoints.route("/projects/<project_id>/evaluations/<model>", methods=["PUT"])
    @rasa_x_scoped("models.evaluations.update", allow_api_token=True)
    async def put_evaluation(request, project_id, model):
        data = _data_service(request).get_training_data(project_id).result
        model_service = _model_service(request)
        model_object = model_service.get_model_by_name(project_id, model)
        if not model_object:
            return error(
                400,
                "NluEvaluationFailed",
                "Could not find requested model '{}'.".format(model),
            )

        evaluation_service = _evaluation_service(request, project_id, "worker")

        try:
            content = await evaluation_service.evaluate(data, model)
            evaluation = evaluation_service.persist_evaluation(
                project_id, model, content
            )
            return response.json(evaluation.as_dict())
        except ClientError as e:
            return error(
                500,
                "NluEvaluationFailed",
                "Failed to create evaluation for model '{}'.".format(model),
                details=e,
            )

    @nlu_endpoints.route(
        "/projects/<project_id>/evaluations/<model>", methods=["GET", "HEAD"]
    )
    @rasa_x_scoped("models.evaluations.get", allow_api_token=True)
    async def get_evaluation(request, project_id, model):
        evaluation_service = _evaluation_service(request, project_id, "worker")
        evaluation = evaluation_service.evaluation_for_model(project_id, model)
        if evaluation:
            return response.json(evaluation)
        else:
            return error(
                404,
                "NluEvaluationNotFound",
                "Could not find evaluation. "
                "A PUT to this endpoint will create an evaluation.",
            )

    @nlu_endpoints.route(
        "/projects/<project_id>/evaluations/<model>", methods=["DELETE"]
    )
    @rasa_x_scoped("models.evaluations.delete", allow_api_token=True)
    async def delete_evaluation(request, project_id, model):
        evaluation_service = _evaluation_service(request, project_id, "worker")
        delete = evaluation_service.delete_evaluation(project_id, model)
        if delete:
            return response.text("", 204)
        else:
            return error(
                404,
                "NluEvaluationNotFound",
                "Could not find evaluation. "
                "A PUT to this endpoint will create an evaluation.",
            )

    @nlu_endpoints.route("/projects/<project_id>/settings", methods=["GET", "HEAD"])
    @rasa_x_scoped("models.settings.get")
    @inject_rasa_x_user()
    async def project_settings(request, project_id, user=None):
        settings_service = SettingsService(request["db_session"])
        stack_config = settings_service.get_config(user["team"], project_id)
        if not stack_config:
            return error(400, "SettingsFailed", "Could not find settings")

        return response.json({"config": dump_yaml(stack_config)})

    @nlu_endpoints.route("/projects/<project_id>/settings", methods=["PUT"])
    @rasa_x_scoped("models.settings.update")
    @inject_rasa_x_user()
    async def save_config(request, project_id, user=None):
        rjs = request.json
        settings_service = SettingsService(request["db_session"])

        try:
            yaml_config = load_yaml(rjs.get("config"))
            settings_service.inspect_config(yaml_config)

        except yaml.error.YAMLError as e:
            return error(
                400,
                "InvalidConfig",
                "Failed to read configuration file.  Error: {}".format(e),
            )
        except InvalidConfigError as e:
            return error(401, "ConfigMissingKeys", "Error: {}".format(e))

        stack_config = settings_service.save_config(
            user["team"], project_id, yaml_config
        )

        if stack_config:
            settings_service.dump_config_in_local_mode(user["team"], project_id)
            return response.json({"config": dump_yaml(yaml_config)})
        else:
            return error(402, "SaveNluConfigFailed", "Could not save nlu config")

    return nlu_endpoints
